local states = require ("states")
local pinballtable = require ("pinballtable")
local renderImages = false
local resetBall = false

bodies = { }
sprites = { }

function love.load()
    
    -- All coordinates in the physics module are divided by this number
    -- and converted to meters, and it creates a convenient way to draw
    -- the objects directly to the screen without the need for
    -- graphics transformations. 
    local pixelsPerMeter = 64
    local allowSleeping = true
    local gravity = 9.81
    love.physics.setMeter (pixelsPerMeter)
    world = love.physics.newWorld (0, gravity * pixelsPerMeter, allowSleeping)
    world:setCallbacks (beginContact)
    world:setContactFilter (contactFilter)
    -- graphics setup
    --sprites.ball = loadSprite ("images/ball.png")
    sprites.background = loadSprite ("images/background.jpg")
    sprites.table = loadSprite ("images/table.png")
    love.graphics.setBackgroundColor(0, 0, 0)
    states:new (states.launch)
    local font = love.graphics.newFont (40)
    love.graphics.setFont (font)
    loadTable (pinballtable)

end

function love.update (dt)

    states:update (dt)

    if (states.current == states.play) then

        for i, ball in ipairs(bodies.balls) do

            -- Limit ball velocity
            local xvel, yvel = ball.body:getLinearVelocity ( )
            xvel = clamp (pinball.ballVelocity.min, xvel, pinball.ballVelocity.max)
            yvel = clamp (pinball.ballVelocity.min, yvel, pinball.ballVelocity.max)
            ball.body:setLinearVelocity (xvel, yvel)
            
            -- Destroy balls marked for destruction
            if (ball.body:getUserData () == "destroy") then
                ball.body:destroy ()
                table.remove (bodies.balls, i)
                -- reassign the ball reference
                bodies.ball = bodies.balls[1]
            end

            
        end

        if love.keyboard.isDown("lshift") then
            for i, flip in pairs(bodies.flippers) do
                if (flip.orientation == "left") then
                    flip.body:applyTorque (-800000)
                end
            end
        end
        if love.keyboard.isDown("rshift") then
            for i, flip in pairs(bodies.flippers) do
                if (flip.orientation == "right") then
                    flip.body:applyTorque (800000)
                end
            end
        end

        world:update (dt)

        if (resetBall == true) then
            resetBall = false
            bodies.ball.fixture:getUserData().onRamp = false
            bodies.ball.body:setPosition (pinballtable.ball.x, pinballtable.ball.y)
            bodies.ball.body:setLinearVelocity (0, 0)
            states:new (states.drained)
        end

    end

end

function love.keypressed (key, isrepeat)

    if (key == " ") then
        resetBall = true
    elseif (key == "b") then
        createBall (pinballtable.ball.x, pinballtable.ball.y)
    end

end

function love.draw ( )

    love.graphics.setColor (255, 255, 255, 255)

    -- Static background
    --love.graphics.draw (sprites.background.image, 0, 0)

    -- Camera follows the ball
    --love.graphics.translate (0, - bodies.ball.body:getY() + 300)

    -- Show the full table
    love.graphics.scale (pinball.drawScale, pinball.drawScale)

    -- Table background
    love.graphics.draw (sprites.table.image, 0, 0)

    -- WALLS
    love.graphics.setLineWidth (6)
    love.graphics.setColor(133, 153, 0)
    for i, wall in pairs(bodies.walls) do
        love.graphics.line(wall.body:getWorldPoints(wall.shape:getPoints()))
    end

    -- RAMPS
    love.graphics.setLineWidth (6)
    love.graphics.setColor(128, 128, 255, 64)
    --love.graphics.setColor(133, 153, 0, 64)
    for i, obj in pairs(bodies.ramps) do
        love.graphics.line(obj.body:getWorldPoints(obj.shape:getPoints()))
    end

    -- BUMPERS
    love.graphics.setLineWidth (2)
    for i, bump in pairs(bodies.bumpers) do
        love.graphics.setColor(42, 161, 152)
        love.graphics.circle("fill", bump.body:getX(), bump.body:getY(), bump.shape:getRadius() * 0.8)
        love.graphics.setColor(108, 113, 196)
        love.graphics.circle("line", bump.body:getX(), bump.body:getY(), bump.shape:getRadius())
    end

    ---- KICKERS
    --love.graphics.setLineWidth (1)
    --for i, kick in pairs(bodies.kickers) do
        --love.graphics.setColor(108, 196, 113)
        --love.graphics.polygon("fill", kick.body:getWorldPoints(kick.shape:getPoints()))
    --end

    ---- TRIGGERS
    --love.graphics.setLineWidth (1)
    --love.graphics.setColor(255, 255, 255, 64)
    --for i, obj in pairs(bodies.triggers) do
        --love.graphics.polygon("line", obj.body:getWorldPoints(obj.shape:getPoints()))
    --end
    
    -- FLIPPERS
    for i, flip in pairs(bodies.flippers) do
        love.graphics.setColor(108, 113, 196)
        love.graphics.polygon("fill", flip.body:getWorldPoints(flip.shape:getPoints()))
        love.graphics.setLineWidth (4)
        love.graphics.setColor(68, 73, 156)
        love.graphics.polygon("line", flip.body:getWorldPoints(flip.shape:getPoints()))
    end

    -- BALL
    love.graphics.setLineWidth (4)
    for k, ball in pairs(bodies.balls) do
        local alpha = ball.fixture:getUserData ().onRamp and 64 or 255
        love.graphics.setColor(238, 232, 213, alpha)
        love.graphics.circle("fill", ball.body:getX(), ball.body:getY(), ball.shape:getRadius())
        love.graphics.setColor(147, 161, 161, alpha)
        love.graphics.circle("line", ball.body:getX(), ball.body:getY(), ball.shape:getRadius())
    end

    if (states.current == states.launch) then

        local text = ""
        if (states.current.timer > 1) then
            text = "GO!"
        elseif (states.current.timer > 0.5) then
            text = "Set"
        elseif (states.current.timer > 0) then
            text = "Ready"
        end
        
        love.graphics.setColor(255, 255, 255)
        love.graphics.printf (text, 0, 900, 600, "center")
        
    elseif (states.current == states.drained) then

    end

end

function loadTable ( )

    pinball.drawScale = love.window.getHeight() / pinballtable.height

    -- Clear the existing table
    if (bodies.all) then
        for k, v in pairs(bodies.all) do
            if (v.fixture) then v.fixture:destroy () end
            if (v.body) then v.body:destroy () end
        end
    end

    bodies = {
        all = { },
        balls = { },
        walls = { },
        ramps = { },
        bumpers = { },
        kickers = { },
        flippers = { },
        triggers = { }
        }

    -- Walls
    for k, v in pairs (pinballtable.walls) do
        createWall (v)
    end

    -- Ramps
    for k, v in pairs (pinballtable.ramps) do
        createWall (v, true)
    end
    
    -- Bumpers
    for k, bump in pairs (pinballtable.bumpers) do
        createBumper (bump.x, bump.y, bump.r / 10, bump.r)
    end

    -- Kickers
    for k, kicker in pairs (pinballtable.kickers) do
        createKickerChain (4, kicker)
    end

    -- Triggers
    for k, trig in pairs (pinballtable.triggers) do
        createTrigger (trig)
    end
    
    -- Anchor flippers to this static body
    local anchorBody = love.physics.newBody(world, 0, 0)
    table.insert (bodies.all, anchorBody)
    
    -- Create the flippers
    for k, flipper in pairs(pinballtable.flippers) do
        createFlipper (flipper.x, flipper.y, flipper.vertices, anchorBody, flipper.orientation)
    end

    -- Create the ball
    createBall (pinballtable.ball.x, pinballtable.ball.y)
    
    -- The Drain
    local drain = { }
    drain.body = love.physics.newBody (world, 650/2, pinballtable.height+150)
    drain.shape = love.physics.newRectangleShape (600, 50)
    drain.fixture = love.physics.newFixture (drain.body, drain.shape)
    drain.fixture:setUserData ({ type = "drain" })
    table.insert (bodies.all, drain)
    
end

function createWall (vertices, isRamp)

    local shell = { }
    shell.body = love.physics.newBody (world, 0, 0)
    shell.shape = love.physics.newChainShape (false, unpack (vertices))
    shell.fixture = love.physics.newFixture (shell.body, shell.shape)
    shell.fixture:setUserData ({ type=isRamp and "ramp" or "wall" })
    table.insert (bodies.all, shell)
    if (isRamp) then
        table.insert (bodies.ramps, shell)
    else
        table.insert (bodies.walls, shell)
    end

end

-- Create a bumper of radius (r) at position (x, y) with kickforce.
-- Good values of kickforce are between 1-2
function createBumper (x, y, kickforce, r)

    local bump = { }
    bump.body = love.physics.newBody (world, x, y, "kinematic")
    bump.shape = love.physics.newCircleShape (r)
    bump.fixture = love.physics.newFixture (bump.body, bump.shape, 1)
    kickforce = clamp (pinball.bumperForces, kickforce)
    bump.fixture:setRestitution (kickforce)
    table.insert (bodies.all, bump)
    table.insert (bodies.bumpers, bump)
    
end

function createKicker (x, y, kickforce, vertices)

    local kick = { }
    kick.body = love.physics.newBody (world, x, y, "kinematic")
    kick.shape = love.physics.newPolygonShape (unpack (vertices))
    kick.fixture = love.physics.newFixture (kick.body, kick.shape, 0)
    kick.fixture:setRestitution (kickforce)
    table.insert (bodies.all, kick)
    table.insert (bodies.kickers, kick)
    
end

--function createSlingshot (x, y, w, h)

    --local obj = { }
    --obj.body = love.physics.newBody (world, x, y, "static")
    --obj.shape = love.physics.newRectangleShape (w, h)
    --obj.fixture = love.physics.newFixture (obj.body, obj.shape, 0)
    --obj.fixture:setUserData ("slingshot")
    --table.insert (bodies.all, obj)
    --table.insert (bodies.slingshots, obj)
    
--end

function createTrigger (def)

    def.isTrigger = true
    local obj = { }
    obj.body = love.physics.newBody (world, def.x, def.y, "static")
    obj.shape = love.physics.newRectangleShape (def.w, def.h)
    obj.fixture = love.physics.newFixture (obj.body, obj.shape, 0)
    obj.fixture:setUserData (def)
    table.insert (bodies.all, obj)
    table.insert (bodies.triggers, obj)
    
end

function createKickerChain (kickforce, vertices)

    local kick = { }
    kick.body = love.physics.newBody (world, x, y, "kinematic")
    kick.shape = love.physics.newChainShape (false, unpack (vertices))
    kick.fixture = love.physics.newFixture (kick.body, kick.shape, 0)
    kick.fixture:setRestitution (kickforce)
    table.insert (bodies.all, kick)
    table.insert (bodies.kickers, kick)
    
end

function createBall (x, y)

    local ball = { }
    ball.body = love.physics.newBody(world, x, y, "dynamic")
    ball.body:setBullet (true)  -- Force high quality collision detection at performance cost
    ball.shape = love.physics.newCircleShape (15)
    ball.fixture = love.physics.newFixture(ball.body, ball.shape, 1)
    ball.fixture:setUserData ({ type="ball", onRamp=false })
    ball.fixture:setRestitution(0.4)
    bodies.ball = ball
    table.insert (bodies.all, ball)
    table.insert (bodies.balls, ball)

end

-- Create a flipper that is controlled by the player.
-- The orientation "left" or "right" determines the rotation limits.
function createFlipper (x, y, vertices, anchorBody, orientation)

    local flip = { }
    flip.body = love.physics.newBody (world, x, y, "dynamic")
    flip.shape = love.physics.newPolygonShape (unpack (vertices))
    flip.fixture = love.physics.newFixture (flip.body, flip.shape, 1)    -- mass is 1
    flip.fixture:setRestitution (0)
    -- Revolute Joint + Motor
    flip.joint = love.physics.newRevoluteJoint (anchorBody, flip.body, x, y, false)
    flip.joint:setMotorSpeed (200)
    flip.joint:setMotorEnabled (true)
    -- Limit movement
    local limitA = orientation == "left" and 5 or 30
    local limitB = orientation == "right" and 5 or 30
    flip.joint:setLimits (-limitA*DEGTORAD, limitB*DEGTORAD)
    flip.joint:setLimitsEnabled (true)
    flip.orientation = orientation
    table.insert (bodies.all, flip)
    table.insert (bodies.flippers, flip)
        
end

-- Separate balls from solids.
-- Returns [ball body], [ball definition], [solid body], [solid definition]
function separateSolids (a, b)

	local aa=a:getUserData() or { }
	local bb=b:getUserData() or { }

    if (aa.type == "ball") then
        return a:getBody(), aa, b:getBody(), bb
    elseif (bb.type == "ball") then
        return b:getBody(), bb, a:getBody(), aa
    end
    
end

-- Filter which objects collide with each other.
function contactFilter (a, b)

    local ball, ballDef, solid, solidDef = separateSolids (a, b)
    
    -- Not a ball collision.
    if (not ball) then
        return true
    end

    -- Affect the game by the trigger's type
    if (solidDef.type == "slingshot" and not ballDef.onRamp) then

        ball:setLinearVelocity (0, -2000)

    elseif (solidDef.type == "wall") then

        return not ballDef.onRamp

    elseif (solidDef.type == "ramp") then

        return ballDef.onRamp

    elseif (solidDef.type == "ramp enter") then

        ballDef.onRamp = true

    elseif (solidDef.type == "ramp exit") then

        ballDef.onRamp = false

    elseif (solidDef.type == "drain") then
    
        if (#bodies.balls == 1) then
            resetBall = true
            ballDef.onRamp = false
        else
            ball:setUserData ("destroy")
        end

        return false
        
    end

    -- Only allow special collisions while the ball is on a ramp.
    -- We do not want to collide with bumpers on a ramp for example.
    if (ballDef.onRamp) then

        --if (solidDef.type == "score") then

        return false

    end

    -- Pass the ball through triggers
    if (solidDef.isTrigger) then
        return false
    else
        return true
    end
    
end

-- Limit value to a range of min and max
function clamp(min, value, max)
    if (type(min) == "table") then
        max = min.max
        min = min.min
    end
    return math.max(min, math.min(max, value))
end

-- Load a sprite and calculate the center point.
function loadSprite (path)

    -- Store sprites as
    --      sprite.image    
    --      sprite.size     (width, height)
    --      sprite.ox       draw offset x
    --      sprite.oy       draw offset y
    local sprite = { }
    sprite.image = love.graphics.newImage (path)
    sprite.size = { sprite.image:getDimensions () }
    sprite.ox = sprite.size[1] / 2
    sprite.oy = sprite.size[2] / 2
    return sprite

end
